self.__BUILD_MANIFEST=function(s,e){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":[s,"static/chunks/932-43254f6ea073238d.js",e,"static/css/fa0655856173cf41.css","static/chunks/pages/index-86b5e49985b35c98.js"],"/_error":["static/chunks/pages/_error-7a92967bea80186d.js"],"/benchmarks":[s,e,"static/css/c9faa5ef13ea61d7.css","static/chunks/pages/benchmarks-1cae3259234af90f.js"],sortedPages:["/","/_app","/_error","/benchmarks"]}}("static/chunks/424-ca7471161a310844.js","static/chunks/541-589309a2c6f59047.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();